#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2017~2999 - cologler <skyoflw@gmail.com>
# ----------
#
# ----------

import typing
from .common import isinstanceof

@isinstanceof.register(type(typing.Union))
def instancecheck(self, obj, **kwargs):
    return isinstanceof(obj, self.__args__)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2017~2999 - cologler <skyoflw@gmail.com>
# ----------
#
# ----------

import typing
from .common import isinstanceof

@isinstanceof.register(typing.Dict)
def instancecheck(self, obj, **kwargs):
    if not isinstance(obj, dict):
        return False
    kt, vt = self.__args__
    return all(isinstanceof(k, kt) for k in obj.keys()) and all(isinstanceof(v, vt) for v in obj.values())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2017~2999 - cologler <skyoflw@gmail.com>
# ----------
#
# ----------

import typing
import functools
import collections.abc as abccol

_INSTANCECHECK = {}
_SUBCLASSCHECK = {}
_SUBCLASSTABLE = {}

null = object()

def py_instancecheck(self, obj, **kwargs):
    return isinstance(obj, self)

def py_subclasscheck(self, obj, **kwargs):
    return issubclass(obj, self)


def _register(d, *args):
    callback = None
    if len(args) == 1:
        cls, = args
    elif len(args) == 2:
        cls, callback = args
    else:
        raise ValueError
    def r(cb):
        d[cls] = cb
        return cb
    if callback is None:
        return r
    else:
        r(callback)
        return callback

def isinstanceof(x, types: tuple, *, typevar_table: typing.Dict[str, type]=None):
    '''
    same as `isinstance()` in python.
    but this function is support the type in `typing`.
    '''
    kwargs = dict(
        typevar_table=typevar_table
    )
    cls = type(types)
    return _INSTANCECHECK.get(cls, py_instancecheck)(types, x, **kwargs)
isinstanceof.register = functools.partial(_register, _INSTANCECHECK)


def issubclassof(x, types, *, typevar_table: typing.Dict[str, type]=None):
    '''
    same as `issubclass()` in python.
    but this function is support the type in `typing`.
    '''
    cls = type(types)
    return _SUBCLASSCHECK.get(cls, py_subclasscheck)(types, x)
issubclassof.register = functools.partial(_register, _SUBCLASSCHECK)

# generic collection check:

_INSTANCECHECK_COLLECTION_TYPE_MAP = {
    typing.List: list,
    typing.Set: set,
    typing.FrozenSet: frozenset,
    typing.Collection: abccol.Collection,
    typing.Iterable: abccol.Iterable,
}

@isinstanceof.register(typing.GenericMeta)
def genericmeta_instancecheck(self, obj, **kwargs):
    gorg = getattr(self, '_gorg', null)
    func = _INSTANCECHECK.get(gorg, null)
    if func is not null:
        return func(self, obj, **kwargs)
    coltype = _INSTANCECHECK_COLLECTION_TYPE_MAP.get(gorg, null)
    if coltype is not null:
        if not isinstance(obj, coltype):
            return False
        typ, = self.__args__
        return all(isinstanceof(x, typ, **kwargs) for x in obj)
    return py_instancecheck(self, obj, **kwargs)

def declare_interface_impl(cls: type, interface: type):
    '''
    declare a type impl a interface.
    '''
    if not isinstance(cls, type):
        raise TypeError
    if not isinstance(interface, type):
        raise TypeError
    table = _SUBCLASSTABLE.get(cls)
    if table is None:
        _SUBCLASSTABLE[cls] = table = set()
    table.add(interface)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2017~2999 - cologler <skyoflw@gmail.com>
# ----------
#
# ----------

from typing import Iterable
from itertools import zip_longest
from .internal.common import isinstanceof, issubclassof

def match(args: (list, tuple), types: Iterable[type]) -> bool:
    '''
    check whether args match types.

    example:
    ``` py
    `match(('', 1), (str, int)) # True
    ```
    '''

    try:
        if len(args) != len(types):
            return False
    except TypeError:
        # object of type 'types' has no len()
        pass

    empty = object()
    for item, typ in zip_longest(args, types, fillvalue=empty):
        if item is empty or typ is empty:
            return False
        if not isinstanceof(item, typ):
            return False
    return True


__all__ = [
    'isinstanceof',
    'issubclassof',
    'match'
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2017~2999 - cologler <skyoflw@gmail.com>
# ----------
#
# ----------

import os
import importlib

for name in os.listdir(os.path.dirname(__file__)):
    if not name.endswith('.py'):
        continue
    if name == '__init__.py':
        continue
    importlib.import_module('.' + name[:-3], __name__)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2017~2999 - cologler <skyoflw@gmail.com>
# ----------
#
# ----------

import typing
from .common import isinstanceof

@isinstanceof.register(typing.TupleMeta)
def instancecheck(self, obj, **kwargs):
    if not isinstance(obj, tuple):
        return False
    return len(obj) == len(self.__args__) and all(isinstance(x, y) for x, y in zip(obj, self.__args__))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2017~2999 - cologler <skyoflw@gmail.com>
# ----------
#
# ----------

import typing
from .common import isinstanceof, issubclassof

@isinstanceof.register(type(typing.Any))
def instancecheck(self, obj, **kwargs):
    return True

@issubclassof.register(type(typing.Any))
def subclasscheck(self, obj, **kwargs):
    return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2017~2999 - cologler <skyoflw@gmail.com>
# ----------
#
# ----------

import typing
import inspect
from .common import isinstanceof, issubclassof

@isinstanceof.register(typing.CallableMeta)
def instancecheck(self, obj, **kwargs):
    if not callable(obj):
        return False
    if self.__args__ is not None:
        *args, ret = self.__args__
        sign = inspect.signature(obj)
        if len(sign.parameters) != len(args):
            return False
        for i, p in enumerate(sign.parameters.values()):
            if p.annotation is not sign.empty:
                if not issubclassof(args[i], p.annotation, **kwargs):
                    return False
        if sign.return_annotation is not sign.empty:
            if not issubclassof(sign.return_annotation, ret, **kwargs):
                return False
    return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2017 - cologler <skyoflw@gmail.com>
# ----------
#
# ----------

import os
from setuptools import setup, find_packages

VERSION = '0.1.2.1'
DESCRIPTION = ''

long_description = None

if os.path.isfile('README.md'):
    with open('README.md') as f:
        long_description = f.read()

long_description = long_description or DESCRIPTION

setup(
    name = 'istype',
    version = VERSION,
    description = DESCRIPTION,
    long_description = long_description or DESCRIPTION,
    classifiers = [],
    keywords = ['python', 'typing'],
    author = 'cologler',
    author_email='skyoflw@gmail.com',
    url = 'https://github.com/Cologler/istype-python',
    license = 'MIT',
    packages = find_packages(),
    include_package_data = True,
    zip_safe = False,
    install_requires = [],
    entry_points = {},
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2017~2999 - cologler <skyoflw@gmail.com>
# ----------
#
# ----------

import typing
from .common import isinstanceof, issubclassof

@isinstanceof.register(typing.TypeVar)
def instancecheck(self, obj, **kwargs):
    args = self.__constraints__
    if args:
        return isinstanceof(obj, args, **kwargs)
    typevar_table = kwargs.get('typevar_table')
    if isinstance(typevar_table, dict):
        typed = typevar_table.get(self.__name__, None)
        if typed is None:
            typevar_table[self.__name__] = type(obj)
            return True
        else:
            if self.__contravariant__:
                return issubclassof(typed, type(obj))
            return isinstanceof(obj, typed, **kwargs)
    if self.__covariant__:
        raise NotImplementedError
    if self.__contravariant__:
        raise NotImplementedError
    return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2017~2999 - cologler <skyoflw@gmail.com>
# ----------
#
# ----------

from .common import isinstanceof, issubclassof

@isinstanceof.register(tuple)
def instancecheck(self, obj, **kwargs):
    return any(isinstanceof(obj, t, **kwargs) for t in self)

@issubclassof.register(tuple)
def subclasscheck(self, obj, **kwargs):
    return any(issubclassof(obj, t) for t in self)
